/*
 * Include header files for all drivers that have been imported from
 * Atmel Software Framework (ASF).
 */
/*
 * Support and FAQ: visit <a href="https://www.microchip.com/support/">Microchip Support</a>
 */
#include <asf.h>

//#define __NVIC_TC3_MC_TEST_ENABLED__

#define CONF_STDIO_USART_MODULE			EDBG_CDC_MODULE
#define CONF_STDIO_MUX_SETTING			EDBG_CDC_SERCOM_MUX_SETTING
#define CONF_STDIO_PINMUX_PAD0			EDBG_CDC_SERCOM_PINMUX_PAD0
#define CONF_STDIO_PINMUX_PAD1			EDBG_CDC_SERCOM_PINMUX_PAD1
#define CONF_STDIO_PINMUX_PAD2			EDBG_CDC_SERCOM_PINMUX_PAD2
#define CONF_STDIO_PINMUX_PAD3			EDBG_CDC_SERCOM_PINMUX_PAD3
#define CONF_STDIO_BAUDRATE				115200

#define PIN_DATA						PIN_PB00


static struct usart_module cdc_uart_module;




/**
 *  Configure UART console.
 */
static void configure_console(void)
{
	struct usart_config usart_conf;

	usart_get_config_defaults(&usart_conf);
	usart_conf.mux_setting = CONF_STDIO_MUX_SETTING;
	usart_conf.pinmux_pad0 = CONF_STDIO_PINMUX_PAD0;
	usart_conf.pinmux_pad1 = CONF_STDIO_PINMUX_PAD1;
	usart_conf.pinmux_pad2 = CONF_STDIO_PINMUX_PAD2;
	usart_conf.pinmux_pad3 = CONF_STDIO_PINMUX_PAD3;
	usart_conf.baudrate    = CONF_STDIO_BAUDRATE;

	stdio_serial_init(&cdc_uart_module, CONF_STDIO_USART_MODULE, &usart_conf);
	usart_enable(&cdc_uart_module);
}




static void TC3_Init8Bits(void)
{
	PM->APBCMASK.reg |= PM_APBCMASK_TC3;

	GCLK->CLKCTRL.reg = (uint16_t)GCLK_CLKCTRL_CLKEN | GCLK_CLKCTRL_GEN_GCLK0 | GCLK_CLKCTRL_ID_TCC2_TC3;

	while(GCLK->STATUS.bit.SYNCBUSY);

	TC3->COUNT8.CTRLA.reg = TC_CTRLA_MODE_COUNT8;

#ifdef __NVIC_TC3_MC_TEST_ENABLED__
	TC3->COUNT8.CTRLA.reg |= TC_CTRLA_WAVEGEN_MFRQ;
#else
	TC3->COUNT8.CTRLA.reg |= TC_CTRLA_WAVEGEN_NFRQ;
#endif

	TC3->COUNT8.CTRLA.reg |= TC_CTRLA_PRESCALER_DIV16;

	// Test code to check TC3 operation
#ifdef __NVIC_TC3_MC_TEST_ENABLED__
	TC3->COUNT8.INTENSET.reg = TC_INTENSET_MC0;
	TC3->COUNT8.CC[0].reg = 3 - 1;

	TC3->COUNT8.CTRLA.reg |= TC_CTRLA_ENABLE;
	while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
#endif
}




#ifdef __NVIC_TC3_MC_TEST_ENABLED__
void TC3_Handler(void)
{
	LED_Toggle(LED0);

	if(TC3->COUNT8.INTFLAG.bit.MC0 == 1)
	{
		TC3->COUNT8.INTFLAG.bit.MC0 = 1;
	}
}
#endif




static uint32_t DHT_read(uint8_t *prh_i, uint8_t *prh_d, uint8_t *ptemp_i, uint8_t *ptemp_d)
{
	uint32_t dht_data = 0;
	uint8_t rh_i, rh_d, temp_i, temp_d;
	volatile uint8_t i, check_sum = 0;
	volatile uint32_t getcount;

	struct port_config pin_conf;
	port_get_config_defaults(&pin_conf);

	//(1)
	pin_conf.direction  = PORT_PIN_DIR_OUTPUT;
	port_pin_set_config(PIN_DATA, &pin_conf);
	port_pin_set_output_level(PIN_DATA, 0);
	delay_ms(20);

	/* Configure LEDs as outputs, turn them off */
	pin_conf.direction  = PORT_PIN_DIR_INPUT;
	pin_conf.input_pull = PORT_PIN_PULL_NONE;
	port_pin_set_config(PIN_DATA, &pin_conf);

	while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
	TC3->COUNT8.COUNT.reg = 0;

	// Before (2), PIN_DATA checks whether DHT11 exists or not.
	while (port_pin_get_input_level(PIN_DATA) == 0)
	{
		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		getcount = TC3->COUNT8.COUNT.reg;
		
		if(getcount >= 255)	// Means there is no DHT11...
		{
			printf("[0] %d\r\n", getcount);
			return 1;
		}
	}

	while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
	TC3->COUNT8.COUNT.reg = 0;

	//(2)
	while (port_pin_get_input_level(PIN_DATA) == 1)
	{
		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		getcount = TC3->COUNT8.COUNT.reg;
		
		if(getcount > 3 * 50)	//50us
		{
			printf("[1] %d\r\n", getcount);
			return 1;
		}
	}

	while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
	TC3->COUNT8.COUNT.reg = 0;

	//(3)
	while (port_pin_get_input_level(PIN_DATA) == 0)
	{
		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		getcount = TC3->COUNT8.COUNT.reg;

		if(getcount > 3 * 90)	// > 90us
		{
			printf("[2] %d\r\n", getcount);
			return 1;
		}
	}

	while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
	TC3->COUNT8.COUNT.reg = 0;

	//(4)
	while (port_pin_get_input_level(PIN_DATA) == 1)
	{
		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		getcount = TC3->COUNT8.COUNT.reg;

		if(getcount > 3 * 90)	// > 90 us
		{
			printf("[4] %d\r\n", getcount);
			return 1;
		}
	}

	for (i = 0; i < 32; i++)
	{
		while (port_pin_get_input_level(PIN_DATA) == 0);	// 50us

		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		TC3->COUNT8.COUNT.reg = 0;

		while (port_pin_get_input_level(PIN_DATA) == 1);

		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		getcount = TC3->COUNT8.COUNT.reg;

		if(getcount >= 3 * 50)
			dht_data |= (0x80000000 >> i);
	}

	for (i = 0; i < 8; i++)
	{
		while (port_pin_get_input_level(PIN_DATA) == 0); // 50us

		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		TC3->COUNT8.COUNT.reg = 0;

		while (port_pin_get_input_level(PIN_DATA) == 1);

		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);
		getcount = TC3->COUNT8.COUNT.reg;

		if(getcount >= 3 * 50)
			check_sum |= (0x80 >> i);
	}

	while (port_pin_get_input_level(PIN_DATA) == 0); // 50us

	rh_i = dht_data >> 24;
	rh_d = (dht_data >> 16) & 0xFF;
	temp_i = (dht_data >> 8) & 0xFF;
	temp_d = dht_data & 0xFF;

	pin_conf.direction  = PORT_PIN_DIR_OUTPUT;
	port_pin_set_config(PIN_DATA, &pin_conf);
	port_pin_set_output_level(PIN_DATA, 1);

	if ((rh_i + rh_d + temp_i + temp_d) == check_sum)
	{
		*prh_i = rh_i;
		*prh_d = rh_d;
		*ptemp_i = temp_i;
		*ptemp_d = temp_d;

		return 0;
	}
	else
	{
		printf("check_sum Failed 0x%X  0x%X\r\n", (rh_i + rh_d + temp_i + temp_d), check_sum);
		return 1;
	}
}




int main (void)
{
	uint8_t rh_i, rh_d, temp_i, temp_d;

	system_init();

	struct port_config pin_conf;
	port_get_config_defaults(&pin_conf);

	/* Configure LEDs as outputs, turn them off */
	pin_conf.direction  = PORT_PIN_DIR_OUTPUT;

	port_pin_set_config(PIN_DATA, &pin_conf);
	port_pin_set_output_level(PIN_DATA, 1);

	configure_console();

	printf("SAMD21 + DHT11 Test => GCLK[0] clock = %d Hz\r\n", system_gclk_gen_get_hz(0));
	printf("core clock = %d Hz\r\n", system_cpu_clock_get_hz());
	printf("APBC clock = %d Hz\r\n", system_apb_clock_get_hz(SYSTEM_CLOCK_APB_APBC));
	printf("SAMD21 TC3 8bits\r\n\n");

	delay_ms(500);

	TC3_Init8Bits();

#ifdef __NVIC_TC3_MC_TEST_ENABLED__
	NVIC_EnableIRQ(TC3_IRQn);
#endif

	printf("TC3 Start\r\n");

	while (1)
	{
#ifndef __NVIC_TC3_MC_TEST_ENABLED__
		TC3->COUNT8.CTRLA.reg |= TC_CTRLA_ENABLE;
		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);

		if(DHT_read(&rh_i, &rh_d, &temp_i, &temp_d) == 0)
			printf("Humidity %d.%d  Temperature %d.%d\r\n", rh_i, rh_d, temp_i, temp_d);
		else
			printf("Error\r\n");

		TC3->COUNT8.CTRLA.reg &= ~TC_CTRLA_ENABLE;
		while(TC3->COUNT8.STATUS.bit.SYNCBUSY == 1);

		LED_Toggle(LED0);
		delay_ms(2000);
#endif
	}
}


